#pragma once

#include <vector>

#include "Core/Objects/interface/Variation.h"
#include "Core/Objects/interface/Weight.h"

#include <TRandom3.h>

namespace DAS {

////////////////////////////////////////////////////////////////////////////////
/// Generic event class for both gen and rec levels, essentially providing
/// access to the weights via `DAS::Uncertainties::Variation`.
class AbstractEvent {
protected:
    virtual std::string_view weight_group () const = 0;

    AbstractEvent () = default;
    virtual ~AbstractEvent () = default;

public:
    Weights weights; //!< e.g. cross section normalisation

    inline double Weight (const Uncertainties::Variation& v) const //!< weight
    {
        if (v.group == weight_group()) {
            const DAS::Weight& w = weights.at(v.index);
            if (w.i == v.bit) return w;
        }
        return weights.front();
    }
};

////////////////////////////////////////////////////////////////////////////////
/// Contains the hard scale. The weights are typically used for the normalisation
/// to the cross section.
struct GenEvent : public AbstractEvent {
    static inline const char * const WeightVar = "GenEvtWgts"; //< Name of gen weights in variations

    float hard_scale; //!< hard scale, corresponding to `pthat` in Pythia 8

    void clear (); //!< to clear for each new event in *n*-tupliser

private:
    std::string_view weight_group () const override { return WeightVar; }
};

////////////////////////////////////////////////////////////////////////////////
/// Contains the data taking information. The weights are typically used
/// for the application of the prescales (in data) or for the PU reweighting (MC).
struct RecEvent : public AbstractEvent {
    static inline const char * const WeightVar = "RecEvtWgts"; //< Name of rec weights in variations

    int runNo, //!< 6-digit run number
        lumi;  //!< lumi section
    unsigned long long evtNo; //!< event number

    void clear (); //!< to clear for each new event in *n*-tupliser

private:
    std::string_view weight_group () const override { return WeightVar; }
};

////////////////////////////////////////////////////////////////////////////////
/// class Trigger
///
/// This class only makes sense for data.
/// It mainly contains the bit of the fired triggers.
/// Originally, it was also supposed to contain the prescale information; the effective luminosties are used instead.
struct Trigger {
    std::vector<bool> Bit; //!< indicates which trigger has fired
    std::vector<int> PreHLT,   //!< HLT prescale
                     PreL1min, //!< L1 min pre-scale
                     PreL1max; //!< L1 max pre-scale

    void clear (); //!< to clear for each new event in *n*-tupliser
};

////////////////////////////////////////////////////////////////////////////////
/// class MET
///
/// Intended mostly for checks, but not used explicitely anywhere in the analysis.
struct MET {
    float Et,    //!< transverse energy
          SumEt, //!< sum of the transverse energies of all the components (jet, leptons, etc.) present in the event
          Pt,    //!< exactly the same as Et
          Phi;   //!< direction of the total transverse momentum
    std::vector<bool> Bit; //!< flags for application of MET filters (see *n*-tupliser config file)

    void clear (); //!< to clear for each new event in *n*-tupliser
};

////////////////////////////////////////////////////////////////////////////////
/// class PileUp
///
/// Contains all relevant information to apply the PU staub sauger and profile reweighting.
struct PileUp {

    static float MBxsec, MBxsecRelUnc;

    float rho;      //!< soft activity (see formula 7.15 in Patrick's thesis)
    int nVtx;       //!< number of vertices in the event
    float trpu;     //!< true pile-up
    int intpu;      //!< in-time pile-up (i.e. from the same bunch crossing)
    std::vector<float> pthats; //!< all hard scales found in PU
    std::vector<unsigned long long> MBevents; //!< event IDs in MB sample

    float GetTrPU (const char v = '0') const; //!< pile-up variations
    float GetInTimePU (const char v = '0', const bool useDefault = true) const; //!< pile-up variations    
    static TRandom3 r3;
    void clear (); //!< to clear for each new event in *n*-tupliser
};

////////////////////////////////////////////////////////////////////////////////
/// class PrimaryVertex
///
/// Contains all relevant information on the leading primary vertex.
struct PrimaryVertex {
    float Rho,  //!< transverse distance to beam axis
          z,    //!< position on beam axis
          chi2, //!< figure of merit of the vertex fit
          ndof; //!< number of degrees of freedom in vertex fit
    bool fake; //!< flag for fake vertices

    float distGen; //!< distance to gen (only relevant in MC)

    void clear (); //!< to clear for each new event in *n*-tupliser
};

}

#if defined(__ROOTCLING__)
#pragma link C++ class DAS::AbstractEvent +;
#pragma link C++ class DAS::GenEvent +;
#pragma link C++ class DAS::RecEvent +;

#pragma link C++ class DAS::Trigger +;

#pragma link C++ class DAS::PrimaryVertex +;

#pragma link C++ class DAS::PileUp +;

#pragma link C++ class DAS::MET +;

#endif
